
"""
Downloads the following:
- Glove vectors
- Stanford Natural Language Inference (SNLI) Corpus
- WordNet 
- CoreNLP tools
"""

import sys
import os
import time
import zipfile
import gzip
import tarfile
import zipfile
import urllib.request

'''
 urllib.urlretrieve
def callbackfunc(blocknum, blocksize, totalsize):
    @blocknum:  already downloaded
    @blocksize: block size
    @totalsize: totle size
'''
def Schedule(blocknum, blocksize, totalsize):
    speed = (blocknum * blocksize) / (time.time() - start_time)
    # speed_str = " Speed: %.2f" % speed
    speed_str = " Speed: %s" % format_size(speed)
    recv_size = blocknum * blocksize

    f = sys.stdout
    pervent = recv_size / totalsize
    percent_str = "%.2f%%" % (pervent * 100)
    n = round(pervent * 50)
    s = ('#' * n).ljust(50, '-')
    f.write(percent_str.ljust(8, ' ') + '[' + s + ']' + speed_str)
    f.flush()
    # time.sleep(0.1)
    f.write('\r')


def format_size(bytes):
    try:
        bytes = float(bytes)
        kb = bytes / 1024
    except:
        print("the input formate is not correct.")
        return "Error"
    if kb >= 1024:
        M = kb / 1024
        if M >= 1024:
            G = M / 1024
            return "%.3fG" % (G)
        else:
            return "%.3fM" % (M)
    else:
        return "%.3fK" % (kb)

def download(url, dirpath):
    print("downloading with the file from %s" %url)
    filename = url.split('/')[-1]
    filepath = os.path.join(dirpath, filename)
    # os.system('wget {} -O {}'.format(url, filepath))  
    urllib.request.urlretrieve(url,filepath,Schedule)
    return filepath

def unzip(filepath):
    print("Extracting: " + filepath)
    dirpath = os.path.dirname(filepath)
    with zipfile.ZipFile(filepath) as zf:
        zf.extractall(dirpath)  
    os.remove(filepath)

# def ungzip(filepath):
#     print("Extracting: " + filepath)
#     dirpath = os.path.dirname(filepath)
#     with tarfile.open(filepath) as zf:
#         zf.extractall(dirpath)
#     os.remove(filepath)

def download_wordvecs(dirpath):
    if os.path.exists(dirpath):
        print('Found Glove vectors - skip')
        return
    else:
        os.makedirs(dirpath)
    url = 'http://www-nlp.stanford.edu/data/glove.840B.300d.zip'
    unzip(download(url, dirpath))

def download_snli(dirpath):
    if os.path.exists(dirpath):
        print('Found SNLI dataset - skip')
        return
    else:
        os.makedirs(dirpath)
    url = 'https://nlp.stanford.edu/projects/snli/snli_1.0.zip'
    unzip(download(url, dirpath))

if __name__ == '__main__':
    base_dir = os.path.dirname(os.path.realpath(__file__))
    snli_dir = os.path.join(base_dir, 'snli')
    wordvec_dir = os.path.join(base_dir, 'glove')
    start_time = time.time()
    download_snli(snli_dir)
    download_wordvecs(wordvec_dir)

