#!/usr/bin/python
import sys
import os
import numpy
import pickle as pkl
import datetime
from collections import OrderedDict

dic = {'entailment': '0', 'neutral': '1', 'contradiction': '2'}

def build_dictionary(filepaths, dst_path,log_file='.\congfig\log',lowercase=False ):
    word_freqs = OrderedDict()
    for filepath in filepaths:
        print_log('Processing ',filepath,file=log_file)
        with open(filepath, 'r') as f:
            for line in f:
                if lowercase:
                    line = line.lower()
                words_in = line.strip().split(' ')   
                for w in words_in:
                    if w not in word_freqs:
                        word_freqs[w] = 0
                    word_freqs[w] += 1

    words = list(word_freqs.keys())   #freqs = list(word_freqs.values())  

    sorted_idx = numpy.argsort(freqs)    #sorted_words = [words[item] for item in sorted_idx[::-1]] 

    worddict = OrderedDict()
    worddict['_PAD_'] = 0  # default, padding
    worddict['_UNK_'] = 1  # out-of-vocabulary
    worddict['_BOS_'] = 2  # begin of sentence token    worddict['_EOS_'] = 3  # end of sentence token

    for ii, ww in enumerate(sorted_words):
        worddict[ww] = ii + 4

    with open(dst_path, 'wb') as f:
        pkl.dump(worddict, f)

    print_log('Dict size: %d'%(len(worddict)), file=log_file)
    print_log('Done', file=log_file)


def build_sequence(filepath, dst_dir,log_file='.\congfig\log'):
    filename = os.path.basename(filepath)
    print_log('the build_sequence stored at %s'%filename, file=log_file)
    len_p = []
    len_h = []
    with open(filepath) as f, \
         open(os.path.join(dst_dir, 'premise_%s'%filename), 'w') as f1, \
         open(os.path.join(dst_dir, 'hypothesis_%s'%filename), 'w') as f2,  \
         open(os.path.join(dst_dir, 'label_%s'%filename), 'w') as f3:
        next(f) # skip the header row
        for line in f:
            sents = line.strip().split('\t')
            if sents[0] is '-':
                continue

            words_in = sents[1].strip().split(' ')
            words_in = [x for x in words_in if x not in ('(',')')]
            f1.write(' '.join(words_in) + '\n')
            len_p.append(len(words_in))

            words_in = sents[2].strip().split(' ')
            words_in = [x for x in words_in if x not in ('(',')')]
            f2.write(' '.join(words_in) + '\n')
            len_h.append(len(words_in))

            f3.write(dic[sents[0]] + '\n')

    print_log('max min len premise:', max(len_p), min(len_p), file=log_file)
    print_log('max min len hypothesis:', max(len_h), min(len_h),file=log_file)

def make_dirs(dirs):
    for d in dirs:
        if not os.path.exists(d):
            os.makedirs(d)

def print_log(*args, **kwargs):  
    if len(kwargs) > 0:
        print(*args, **kwargs)
    return None

if __name__ == '__main__':
    base_dir = os.path.dirname(os.path.realpath(__file__))
    logs = os.path.join(base_dir, 'logs')
    if not os.path.exists(logs):
        os.mkdir(logs)
    dt = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
    log_path = './logs/log.{}'.format(dt)
    log = open(log_path, 'w', encoding='UTF-8')

    print_log('=' * 80, file=log)
    print_log('Preprocessing snli_1.0 dataset', file=log)
    print_log('=' * 80, file=log)

    dst_dir = os.path.join(base_dir, 'word_sequence')
    snli_dir = os.path.join(base_dir, 'snli/snli_1.0')
    make_dirs([dst_dir])

    build_sequence(os.path.join(snli_dir, 'snli_1.0_dev.txt'), dst_dir,log)
    build_sequence(os.path.join(snli_dir, 'snli_1.0_test.txt'), dst_dir,log)
    build_sequence(os.path.join(snli_dir, 'snli_1.0_train.txt'), dst_dir,log)

    build_dictionary([os.path.join(dst_dir, 'premise_snli_1.0_train.txt'), 
                      os.path.join(dst_dir, 'hypothesis_snli_1.0_train.txt')], 
                      os.path.join(dst_dir, 'vocab_cased.pkl'),log)

    make_dirs([os.path.join(os.path.dirname(base_dir), 'models')])