# -*- coding: utf-8 -*
import os
import numpy
import theano
import pickle
import subprocess
theano.config.floatX='float32'

def datafold(textFile):
    x, y, z = pickle.load(open(textFile, "rb"))
    return x, y, z

def dicfold(textFile):
    word2idx, label2idx = pickle.load(open(textFile, "rb"))
    # print(type(word2idx))
    # print(type(label2idx))
    return word2idx, label2idx

def get_model_size(params):
    total_size = 0
    for k, v in params.items():
        total_size += v.size
    return total_size

def get_minibatches_idx(n, minibatch_size, shuffle=False):
    
    idx_list = numpy.arange(n,dtype="int64")
    # idx_list = numpy.arange(n)

    if shuffle:
        numpy.random.shuffle(idx_list)

    minibatches = []
    minibatch_start = 0
    for i in range(n // minibatch_size):
        minibatches.append(idx_list[minibatch_start: minibatch_start + minibatch_size])
        minibatch_start += minibatch_size

    if (minibatch_start != n):
        # Make a minibatch out of what is left
        minibatches.append(idx_list[minibatch_start:])

    return zip(range(len(minibatches)), minibatches)

def split_idx(nsentences, valid_rate):
    
    idx_list = numpy.arange(nsentences, dtype="int64")
    # idx_list = numpy.arange(nsentences)
    numpy.random.shuffle(idx_list)
    return idx_list[:int(nsentences * valid_rate)], idx_list[int(nsentences * valid_rate):]


def str2list(s):
    alphabet = " abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789,;.!?:'\"/\\|_@#$%^&*~`+-=<>()[]{}"
    l = len(s)
    ans = []
    for i in range(0, l):
        a = alphabet.find(s[i])   
        if a >= 0:
            ans.append(a)
        else:
            ans.append(0)
            #print(s[i])
    return ans


def prepare_data(seqs, worddicts_r,maxlen = None):
    
    lengths = [len(s) for s in seqs]

    if maxlen:
        new_seqs = []
        new_lengths = []
        for l, s in list(zip(lengths, seqs)):
            if l < maxlen:
                new_seqs.append(s)
                new_lengths.append(l)
            else:
                new_seqs.append(s[: maxlen])
                new_lengths.append(maxlen)
        seqs = new_seqs
        lengths = new_lengths

        if len(lengths) < 1:
            return None, None

    max_char_len = 0
    seqs_char = []
    l_seqs_char = []

    for s_x in seqs:
        # print(s_x,type(s_x))
        temp_seqs_char = []
        temp_l_seqs_char = []
        for w_x in s_x:
            # print(w_x,type(w_x))
            word = worddicts_r[w_x]  
            # print(word,type(word))
            word_list = str2list(word)
            # print(word_list,type(word_list))
            l_word_list = len(word_list)
            temp_seqs_char.append(word_list)
            temp_l_seqs_char.append(l_word_list)
            if l_word_list >= max_char_len:
                max_char_len = l_word_list

        seqs_char.append(temp_seqs_char)
        l_seqs_char.append(temp_l_seqs_char)

    n_samples = len(seqs)
    max_len = max(lengths)

    x = numpy.zeros((max_len, n_samples)).astype('int64')
    x_mask = numpy.zeros((max_len, n_samples)).astype('float32')

    char_x = numpy.zeros((max_len, n_samples, max_char_len)).astype('int64')
    char_x_mask = numpy.zeros((max_len, n_samples, max_char_len)).astype('float32')

    for idx, s in enumerate(seqs):
        x[:lengths[idx], idx] = s
        x_mask[:lengths[idx], idx] = 1.

        for j in range(0, lengths[idx]):
            char_x[j, idx, :l_seqs_char[idx][j]] = seqs_char[idx][j]
            char_x_mask[j, idx, :l_seqs_char[idx][j]] = 1.

    return x, x_mask,char_x, char_x_mask

def getA(answers, results):           
    total = 0.
    right = 0.
    # total += len(answers)
    for i in range(len(answers)):
        if answers[i] == results[i]:
            right += 1
        total += 1
    return right / total

def getP(answers, results, tag):
    total = 0.
    right = 0.
    for i in range(len(results)):
        if results[i] == tag:
            total += 1
        if answers[i] == tag and results[i] == tag:
            right += 1
    print("P total: %d" %total)
    print("P right: %d" %right)
    return 0. if total == 0 else right / total

def getR(answers, results, tag):       
    total = 0.
    right = 0.
    for i in range(len(answers)):
        if answers[i] == tag:
            total += 1
        if answers[i] == tag and results[i] == tag:
            right += 1
    print("R total: %d" %total)
    print("R right: %d" %right)
    return 0. if total == 0 else right / total

def getF1(precision, recall):         
    return 0. if (precision + recall) == 0 else 2. * precision * recall / (precision + recall)

def evaluate(answers, results, filename):     
    print("answers.size: %d" %len(answers))
    print("results.size: %d" %len(results))
    f = open(filename, "w")
    for p in results:
        f.write(str(p) + "\n")
    f.close()

    pos_P = getP(answers, results, 1)
    pos_R = getR(answers, results, 1)
    pos_F1 = getF1(pos_P, pos_R)

    neg_P = getP(answers, results, 0)
    neg_R = getR(answers, results, 0)
    neg_F1 = getF1(neg_P, neg_R)

    accuracy = getA(answers, results)

    return pos_P, pos_R, pos_F1, neg_P, neg_R, neg_F1, accuracy

def save_score(score, filename):
    with open(filename, "w") as f:
        for s in score:
            f.write("%f\n" %s)

def save_label(label, filename):
    with open(filename, "w") as f:
        for l in label:
            f.write("%d\n" %l)

def pred_acc(f_pred, pred_file, prepare_data,worddicts_r, data, kf_data, options):   
    data_acc = 0
    n_done = 0
    labels = []
    preds = []

    for _, data_index in kf_data:
        # n_done +=len(data_index)
        # print(n_done)
        x = [data[0][t] for t in data_index]
        y = [data[1][t] for t in data_index]
        z = [data[2][t] for t in data_index]

        x_, mask_x_, char_x_, char_x_mask_ = prepare_data(x,worddicts_r)
        y_, mask_y_, char_y_, char_y_mask_ = prepare_data(y,worddicts_r)
        z_ = numpy.array(z)

        labels.extend(z_)
        preds.extend(f_pred(x_, y_,mask_x_,mask_y_, char_x_, char_y_, char_x_mask_ ,char_y_mask_))
        # preds = f_pred(x_, y_, mask_x_, mask_y_)
        # data_acc += (preds==z).sum()
    # data_acc = 1.0 *data_acc /n_done
    data_acc = getA(labels, preds)

    with open(pred_file, "w") as f:
        for pred in preds:
            f.write("%d\n" %pred)

    return data_acc

def pred_probs(f_log_prob, prepare_data,worddicts_r, data, data_kf, options):
    probs = []

    for _, data_index in data_kf:
        x = [data[0][t] for t in data_index]
        y = [data[1][t] for t in data_index]
        z = [data[2][t] for t in data_index]

        x_, mask_x_ ,char_x_, char_x_mask_= prepare_data(x,worddicts_r)
        y_, mask_y_ ,char_y_, char_y_mask_= prepare_data(y,worddicts_r)
        z_ = numpy.array(z)

        pprobs = f_log_prob(x_, y_,mask_x_, mask_y_,char_x_,char_y_,  char_x_mask_, char_y_mask_,z_)
        probs.append(pprobs)

    return numpy.array(probs).mean()


if __name__ == '__main__':
    # x,y,z = datafold('../data/workshop/snli_1.0_train.pkl')
    # print(x,type(x))
    # print("-"*30)
    # print(y,type(y))
    # print("-" * 30)
    # print(z,type(z))
    # print("-"*30)
	x,y = dicfold("../data/workshop/snli_1.0_dic.pkl")
	print(x,type(x))
	print("-" * 30)
	worddicts_r = dict()
	for kk, vv in x.items():
		worddicts_r[vv] = kk
	print(worddicts_r,type(worddicts_r))
	print("-" * 30)
	print(y,type(y))