# -*- coding: utf-8 -*
from collections import OrderedDict
import pickle as pkl
# from mr_phrase.construct_nn.Paraphrase import process
import os
import numpy
import theano
import theano.tensor as tensor
from theano import config
from theano.sandbox.rng_mrg import MRG_RandomStreams as RandomStreams

import sys
sys.path.append(".")
from library import lib,optimization
theano.config.floatX = 'float32'
# config.optimizer = 'fast_compile'

optimizer = {"adam": optimization.adam, "adadelta": optimization.adadelta}

# make prefix-appended name
def _p(pp, name):
	return '%s_%s' % (pp, name)

# push parameters to Theano shared variables
def zipp(params, tparams):
	for kk, vv in params.items():
		tparams[kk].set_value(vv)

# push parameters to Theano shared variables
def unzip(zipped):
	new_params = OrderedDict()
	for kk, vv in zipped.items():
		new_params[kk] = vv.get_value()
	return new_params

# get the list of parameters: Note that tparams must be OrderedDict
def itemlist(tparams):
	return [vv for kk, vv in tparams.items()]

def load_params(path, params):
	pp = numpy.load(path)
	for kk, vv in params.items():
		if kk not in pp:
			raise Warning("%s is not in the archive" % kk)
		params[kk] = pp[kk]

	return params

# feedforward layer: affine transformation + point-wise nonlinearity
def param_init_multihead_attention(options, params,
								   num_heads=3,
								   prefix='multihead_attention', nin=None, nout=None, **kwargs):
	if nin is None:
		nin = options.dim
	if nout is None:
		nout = options.dim
	params[_p(prefix, 'W1')] = lib.norm_weight(nin, nout*num_heads, scale=0.01, ortho=False)
	params[_p(prefix, 'W2')] = lib.norm_weight(nout, nin*num_heads, scale=0.01, ortho=False)
	params[_p(prefix, 'b1')] = numpy.zeros((nout*num_heads,)).astype('float32')
	params[_p(prefix, 'b2')] = numpy.zeros((nin*num_heads,)).astype('float32')

	return params

def init_tparams(params):
	tparams = OrderedDict()
	print("the initial parameters: ")
	for kk, pp in params.items():
		tparams[kk] = theano.shared(params[kk], name=kk)
		print(kk, pp.shape)
	return tparams

# def init_data_params(options, worddicts):
#
# 	return params

def concatenate(tensor_list, axis=0):
    concat_size = sum(tt.shape[axis] for tt in tensor_list)

    output_shape = ()
    for k in range(axis):
        output_shape += (tensor_list[0].shape[k],)
    output_shape += (concat_size,)
    for k in range(axis + 1, tensor_list[0].ndim):
        output_shape += (tensor_list[0].shape[k],)

    out = tensor.zeros(output_shape)
    offset = 0
    for tt in tensor_list:
        indices = ()
        for k in range(axis):
            indices += (slice(None),)
        indices += (slice(offset, offset + tt.shape[axis]),)
        for k in range(axis + 1, tensor_list[0].ndim):
            indices += (slice(None),)

        out = tensor.set_subtensor(out[indices], tt)
        offset += tt.shape[axis]

    return out

def multihead_attention(tparams, options,
                        inputs, mask, num_heads=3,
                        prefix='multihead_attention', **kwargs):
    # inputs (nstep_x1, batch, 600)
    # mask  (nstep_x1, batch)
    # mask_dis  (batch, nstep_x1, nstep_x1)
    step = inputs.shape[0]
    batch = inputs.shape[1]
    nout = tparams[_p(prefix, 'W1')].shape[1]//num_heads
    nin = tparams[_p(prefix, 'W1')].shape[0]
    # attention = []

    multiatt = lib.ReLU(tensor.dot(inputs, tparams[_p(prefix, 'W1')]) + tparams[_p(prefix, 'b1')])
    multiatt = multiatt.reshape([step, batch, num_heads, nout])
    multiatt = multiatt.dimshuffle(2,0,1,3)
    multiatt = tensor.batched_tensordot(multiatt, tparams[_p(prefix, 'W2')].reshape([nout, num_heads, nin]).dimshuffle(1,0,2), axes=[3,1])
    multiatt += tparams[_p(prefix, 'b2')].reshape([num_heads, nin])[:,None,None,:]
    multiatt = multiatt.dimshuffle(1,2,3,0)

    multiatt = tensor.exp(multiatt - multiatt.max(0, keepdims=True))
    multiatt = multiatt * mask[:,:,None,None]
    multiatt = multiatt / (multiatt.sum(0, keepdims=True)  + 1e-8)
    outputs = (inputs.dimshuffle(0,1,2,'x') * multiatt).sum(3)

    return outputs, multiatt

def split(x, num, axis=1):
    out = []
    dim = x.shape[axis]//num
    if axis == 1:
        for n in range(num):
            out.append(x[:, n*dim:(n+1)*dim])
    return out

""" 
Builds the entire computational graph used for training
"""
def build_model(options):


	# params = OrderedDict()
	trng = RandomStreams(options.seed)

	use_noise = theano.shared(lib.numpy_floatX(0.))

	options.char_k_rows = [1, 3, 5]

	params = OrderedDict()


	params['Wemb'] = lib.norm_weight(nin=options.n_words, nout=options.dim_word)
	if options.file_emb:
		print('load emb from ' + options.file_emb)
		with open(options.file_emb, 'r', encoding='UTF-8') as f:
			for line in f:
				tmp = line.split()
				word = tmp[0]
				vector = tmp[1:]
				if word in options.dic and options.dic[word] < options.n_words:
					if len(vector) != 300:
						continue
					else:
						params['Wemb'][options.dic[word], :] = vector


	params['Charemb'] = lib.norm_weight(nin=options.l_alphabet + 1, nout=options.dim_char_emb)
	for char_k_rows in options.char_k_rows:
		w_shp = (options.char_nout, 1, char_k_rows, options.char_k_cols)
		w_bound = numpy.sqrt(3 * char_k_rows * options.char_k_cols)
		params[_p("filter",str(char_k_rows))] = numpy.random.uniform(low=-1.0 / w_bound, high=1.0 / w_bound,size=w_shp).astype(config.floatX)

	dim_emb = options.dim_word + 3 * options.char_nout

	lstm_fw_input_encoders = []
	lstm_bw_input_encoders = []


	for l in range(options.encoder_layers):
		if l == 0:
			lstm_fw_input_encoder = lib.LSTM_layer(prefix=_p("encoder", str(l + 1)), dim_in=dim_emb,
												   dim_out=options.dim)
			lstm_fw_input_encoder.init_params(options, params)
			lstm_fw_input_encoders.append(lstm_fw_input_encoder)

			lstm_bw_input_encoder = lib.LSTM_layer(prefix=_p("encoder_r", str(l + 1)), dim_in=dim_emb,
												   dim_out=options.dim)
			lstm_bw_input_encoder.init_params(options, params)
			lstm_bw_input_encoders.append(lstm_bw_input_encoder)

		else:
			lstm_fw_input_encoder = lib.LSTM_layer(prefix=_p("encoder", str(l + 1)), dim_in=options.dim * 2 + dim_emb,
												   dim_out=options.dim)
			lstm_fw_input_encoder.init_params(options, params)
			lstm_fw_input_encoders.append(lstm_fw_input_encoder)

			lstm_bw_input_encoder = lib.LSTM_layer(prefix=_p("encoder_r", str(l + 1)), dim_in=options.dim * 2 + dim_emb,
												   dim_out=options.dim)
			lstm_bw_input_encoder.init_params(options, params)
			lstm_bw_input_encoders.append(lstm_bw_input_encoder)

	params = param_init_multihead_attention(options, params,
											num_heads=options.num_heads,
											prefix='multihead_attention',
											nin=options.dim * 2,
											nout=options.dim)


	for i in range(options.num_block):
		external_att_W = lib.norm_weight(options.dim*2, options.dim * 2)
		params[_p("external_att_W", str(i))] = external_att_W
	
		external_att_W_x = lib.norm_weight(1, options.dim * 2).flatten()
		params[_p("external_att_W_x", str(i))] = external_att_W_x
	
		external_att_W_y = lib.norm_weight(1, options.dim * 2).flatten()
		params[_p("external_att_W_y", str(i))] = external_att_W_y
	
	mlp_layer_1s = []
	mlp_layer_1fs = []
	mlp_layer_2s = []
	lstm_fw_input_2s = []
	lstm_bw_input_2s = []

	for i in range(options.num_block):
		mlp_layer_1 = lib.MLP_layer(prefix = _p("mlp_layer_1", str(i)), dim_in = options.dim*2* 4, dim_out = options.dim_mlp_1)
		mlp_layer_1.init_params(options, params)
		mlp_layer_1s.append(mlp_layer_1)

		mlp_layer_1f = lib.MLP_layer(prefix=_p("mlp_layer_1f", str(i)), dim_in=options.dim_mlp_1, dim_out=options.dim*2)
		mlp_layer_1f.init_params(options, params)
		mlp_layer_1fs.append(mlp_layer_1f)

		mlp_layer_2 = lib.MLP_layer(prefix = _p("mlp_layer_2", str(i)), dim_in = options.dim* 2 * 4, dim_out = options.dim_mlp_2)
		mlp_layer_2.init_params(options, params)
		mlp_layer_2s.append(mlp_layer_2)
		
		lstm_fw_input_2 = lib.LSTM_layer(prefix = _p("lstm_fw_input_2", str(i)), dim_in = options.dim_mlp_2, dim_out = options.dim_lstm_input_0)
		lstm_fw_input_2.init_params(options, params)
		lstm_fw_input_2s.append(lstm_fw_input_2)

		lstm_bw_input_2 = lib.LSTM_layer(prefix = _p("lstm_bw_input_2", str(i)), dim_in = options.dim_mlp_2, dim_out = options.dim_lstm_input_0)
		lstm_bw_input_2.init_params(options, params)
		lstm_bw_input_2s.append(lstm_bw_input_2)

	mlp_layer_3 = lib.MLP_layer(prefix = "dense", dim_in = options.dim_lstm_input_0 * 2 * 4, dim_out = options.dim_dense)
	mlp_layer_3.init_params(options, params)
	
	mlp_layer_4 = lib.MLP_layer(prefix = "predicator", dim_in = options.dim_dense, dim_out = options.num_class)
	mlp_layer_4.init_params(options, params)


	if options.reload_model:
		load_params(os.path.join(options.folder, options.reload_model), params)
	
	tparams = init_tparams(params)

	
	x1 = tensor.matrix('x1', dtype="int64")
	x1_mask = tensor.matrix('x1_mask', dtype="float32")
	x2 = tensor.matrix('x2', dtype="int64")
	x2_mask = tensor.matrix('x2_mask', dtype="float32")
	z = tensor.vector('z', dtype="int64")

	xr1 = x1[::-1]
	xr1_mask = x1_mask[::-1]
	xr2 = x2[::-1]
	xr2_mask = x2_mask[::-1]

	n_timesteps_x1 = x1.shape[0]
	n_timesteps_x2 = x2.shape[0]
	n_samples = x1.shape[1]

	char_x1 = theano.tensor.tensor3('char_x1', dtype='int64')
	char_x1_mask = theano.tensor.tensor3('char_x1_mask', dtype='float32')
	char_x2 = theano.tensor.tensor3('char_x2', dtype='int64')
	char_x2_mask = theano.tensor.tensor3('char_x2_mask', dtype='float32')

	emb_char1 = tparams['Charemb'][char_x1.flatten()].reshape([n_timesteps_x1, n_samples, char_x1.shape[2], options.dim_char_emb])
	emb_char1 = emb_char1 * char_x1_mask[:, :, :, None]
	emb_char_inp1 = emb_char1.reshape([n_timesteps_x1 * n_samples, 1, char_x1.shape[2], options.dim_char_emb])
	emb_char1s = []
	for num in options.char_k_rows:
		emb_char1 = tensor.nnet.conv2d(emb_char_inp1, tparams[_p("filter",str(num))], border_mode='valid')
		emb_char1 = tensor.nnet.nnet.relu(emb_char1)
		emb_char1 = emb_char1.reshape([n_timesteps_x1 * n_samples, options.char_nout, emb_char1.shape[2]])
		emb_char1 = emb_char1.max(2)
		emb_char1 = emb_char1.reshape([n_timesteps_x1, n_samples, options.char_nout])
		emb_char1s.append(emb_char1)

	emb_char1 = concatenate(emb_char1s, axis=2)

	emb_char2 = tparams['Charemb'][char_x2.flatten()].reshape(
		[n_timesteps_x2, n_samples, char_x2.shape[2], options.dim_char_emb])  #()
	emb_char2 = emb_char2 * char_x2_mask[:, :, :, None]
	emb_char_inp2 = emb_char2.reshape([n_timesteps_x2 * n_samples, 1, char_x2.shape[2], options.dim_char_emb])
	emb_char2s = []
	for num in options.char_k_rows:
		emb_char2 = tensor.nnet.conv2d(emb_char_inp2, tparams[_p("filter",str(num))], border_mode='valid')
		emb_char2 = tensor.nnet.nnet.relu(emb_char2)
		emb_char2 = emb_char2.reshape([n_timesteps_x2 * n_samples, options.char_nout, emb_char2.shape[2]])
		emb_char2 = emb_char2.max(2)
		emb_char2 = emb_char2.reshape([n_timesteps_x2, n_samples, options.char_nout])
		emb_char2s.append(emb_char2)

	emb_char2 = concatenate(emb_char2s, axis=2)

	# word embedding
	emb1 = tparams['Wemb'][x1.flatten()].reshape([n_timesteps_x1, n_samples, options.dim_word])
	emb1 = concatenate([emb1, emb_char1], axis=2)
	if options.use_dropout:
		emb1 = lib.dropout_data_layer(emb1, use_noise, trng)

	emb2 = tparams['Wemb'][x2.flatten()].reshape([n_timesteps_x2, n_samples, options.dim_word])
	emb2 = concatenate([emb2, emb_char2], axis=2)
	if options.use_dropout:
		emb2 = lib.dropout_data_layer(emb2, use_noise, trng)

	for l in range(options.encoder_layers):
		if l == 0:
			ctx1 = emb1
			ctx2 = emb2
		else:
			ctx1 = concatenate([ctx1, emb1], axis=2)
			ctx2 = concatenate([ctx2, emb2], axis=2)

		ctxr1 = ctx1[::-1]
		ctxr2 = ctx2[::-1]
		proj1 = lstm_fw_input_encoders[l].build_rep(tparams,ctx1,options,mask=x1_mask)
		projr1 = lstm_bw_input_encoders[l].build_rep(tparams,ctxr1,options,mask=xr1_mask)

		proj2 = lstm_fw_input_encoders[l].build_rep(tparams, ctx2, options, mask=x2_mask)
		projr2 = lstm_bw_input_encoders[l].build_rep(tparams, ctxr2, options, mask=xr2_mask)

		ctx1 = concatenate([proj1[0], projr1[0][::-1]], axis=proj1[0].ndim - 1)
		ctx2 = concatenate([proj2[0], projr2[0][::-1]], axis=proj2[0].ndim - 1)


	att_1, multiatt_1 = multihead_attention(tparams, options, ctx1, x1_mask, num_heads=options.num_heads)
	att_2, multiatt_2 = multihead_attention(tparams, options, ctx2, x2_mask, num_heads=options.num_heads)

	rep1_origin = att_1
	rep2_origin = att_2

	rep1 = att_1.reshape([att_1.shape[0], att_1.shape[1] * att_1.shape[2]])
	rep2 = att_2.reshape([att_2.shape[0], att_2.shape[1] * att_2.shape[2]])

	opt_ret = dict()
	opt_ret['rep1'] = rep1
	opt_ret['rep2'] = rep2
	opt_ret['rep1_origin'] = rep1_origin
	opt_ret['rep2_origin'] = rep2_origin
	opt_ret['multiatt_1'] = multiatt_1
	opt_ret['multiatt_2'] = multiatt_2

	input_xs = [rep1_origin ]
	input_ys = [rep2_origin]

	for i in range(options.num_block):
		input_x_0 = input_xs[-1]
		input_y_0 = input_ys[-1]

		external_weight_matrix = tensor.dot(input_x_0.dimshuffle(1,0,2), tparams[_p("external_att_W", str(i))])
		external_weight_matrix = tensor.batched_dot(external_weight_matrix, input_y_0.dimshuffle(1,2,0))
		external_weight_vec_x = tensor.dot(input_x_0, tparams[_p("external_att_W_x", str(i))]).dimshuffle(1, 0)
		external_weight_vec_y = tensor.dot(input_y_0, tparams[_p("external_att_W_y", str(i))]).dimshuffle(1, 0)
		external_weight_matrix = external_weight_matrix + external_weight_vec_x[:, :, None] + external_weight_vec_y[:, None, :]

		external_weight_matrix_x = tensor.exp(external_weight_matrix - external_weight_matrix.max(axis = 1, keepdims = True)).dimshuffle(1, 2, 0)
		external_weight_matrix_y = tensor.exp(external_weight_matrix - external_weight_matrix.max(axis = 2, keepdims = True)).dimshuffle(1, 2, 0)
		external_weight_matrix_x = external_weight_matrix_x * x1_mask[:, None, :]
		external_weight_matrix_y = external_weight_matrix_y * x2_mask[None, :, :]
		external_alpha = external_weight_matrix_x / (external_weight_matrix_x.sum(axis = 0, keepdims = True)+ 1e-8)
		external_beta = external_weight_matrix_y / (external_weight_matrix_y.sum(axis = 1, keepdims = True)+ 1e-8)
	
		input_y_external_att = (input_x_0.dimshuffle(0, 'x', 1, 2) * external_alpha.dimshuffle(0, 1, 2, 'x')).sum(0)
		input_x_external_att = (input_y_0.dimshuffle('x', 0, 1, 2) * external_beta.dimshuffle(0, 1, 2, 'x')).sum(1)

		
		input_x_0_cat = tensor.concatenate([input_x_0, input_x_external_att, input_x_0 - input_x_external_att, input_x_0 * input_x_external_att], axis = -1)
		input_y_0_cat = tensor.concatenate([input_y_0, input_y_external_att, input_y_0 - input_y_external_att, input_y_0 * input_y_external_att], axis = -1)

		input_x_0_cat_dp = lib.dropout_layer(input_x_0_cat, use_noise, trng, options)
		input_y_0_cat_dp = lib.dropout_layer(input_y_0_cat, use_noise, trng, options)

		input_x_mlp_1 = tensor.nnet.relu(mlp_layer_1s[i].build(tparams, input_x_0_cat_dp, options))
		input_y_mlp_1 = tensor.nnet.relu(mlp_layer_1s[i].build(tparams, input_y_0_cat_dp, options))
		
		input_x_mlp_1_dp = lib.dropout_layer(input_x_mlp_1, use_noise, trng, options)
		input_y_mlp_1_dp = lib.dropout_layer(input_y_mlp_1, use_noise, trng, options)

		input_x_1 = mlp_layer_1fs[i].build(tparams,input_x_mlp_1_dp,options) * x1_mask[:, :, None]
		input_x_1 = input_x_0 + input_x_1

		input_y_1 = mlp_layer_1fs[i].build(tparams, input_y_mlp_1_dp, options) * x2_mask[:, :, None]
		input_y_1 = input_y_0 + input_y_1

		
		internal_weight_matrix_x = tensor.batched_dot(input_x_1.dimshuffle(1, 0, 2), input_x_1.dimshuffle(1, 2, 0))
		internal_weight_matrix_x = tensor.exp(internal_weight_matrix_x - internal_weight_matrix_x.max(axis = 1, keepdims = True)).dimshuffle(1, 2, 0)
		internal_weight_matrix_x = internal_weight_matrix_x * x1_mask[:, None, :]
		internal_alpha = internal_weight_matrix_x / (internal_weight_matrix_x.sum(axis = 0, keepdims = True)+ 1e-8)
		
		internal_weight_matrix_y = tensor.batched_dot(input_y_1.dimshuffle(1, 0, 2), input_y_1.dimshuffle(1, 2, 0))
		internal_weight_matrix_y = tensor.exp(internal_weight_matrix_y - internal_weight_matrix_y.max(axis = 1, keepdims = True)).dimshuffle(1, 2, 0)
		internal_weight_matrix_y = internal_weight_matrix_y * x2_mask[:, None, :]
		internal_beta = internal_weight_matrix_y / (internal_weight_matrix_y.sum(axis = 0, keepdims = True)+ 1e-8)
		
		input_x_internal_att = (input_x_1.dimshuffle(0, 'x', 1, 2) * internal_alpha.dimshuffle(0, 1, 2, 'x')).sum(0)
		input_y_internal_att = (input_y_1.dimshuffle(0, 'x', 1, 2) * internal_beta.dimshuffle(0, 1, 2, 'x')).sum(0)

		
		input_x_1_cat = tensor.concatenate([input_x_1, input_x_internal_att, input_x_1 - input_x_internal_att, input_x_1 * input_x_internal_att], axis = -1)
		input_y_1_cat = tensor.concatenate([input_y_1, input_y_internal_att, input_y_1 - input_y_internal_att, input_y_1 * input_y_internal_att], axis = -1)

		input_x_1_cat_dp = lib.dropout_layer(input_x_1_cat, use_noise, trng, options)
		input_y_1_cat_dp = lib.dropout_layer(input_y_1_cat, use_noise, trng, options)

		input_x_mlp_2 = tensor.nnet.relu(mlp_layer_2s[i].build(tparams, input_x_1_cat_dp, options))
		input_y_mlp_2 = tensor.nnet.relu(mlp_layer_2s[i].build(tparams, input_y_1_cat_dp, options))
		
		input_x_mlp_2_dp = lib.dropout_layer(input_x_mlp_2, use_noise, trng, options)
		input_y_mlp_2_dp = lib.dropout_layer(input_y_mlp_2, use_noise, trng, options)
		
		input_fw_x_2 = lstm_fw_input_2s[i].build(tparams, input_x_mlp_2_dp, options, x1_mask)
		input_bw_x_2 = lib.reverse(lstm_bw_input_2s[i].build(tparams, lib.reverse(input_x_mlp_2_dp), options, lib.reverse(x1_mask)))
		input_x_2 = tensor.concatenate([input_fw_x_2, input_bw_x_2], axis = -1) * x1_mask[:, :, None]

		input_fw_y_2 = lstm_fw_input_2s[i].build(tparams, input_y_mlp_2_dp, options,x2_mask)
		input_bw_y_2 = lib.reverse(lstm_bw_input_2s[i].build(tparams, lib.reverse(input_y_mlp_2_dp), options, lib.reverse(x2_mask)))
		input_y_2 = tensor.concatenate([input_fw_y_2, input_bw_y_2], axis=-1) * x2_mask[:, :, None]
		
		input_xs.append(input_x_2)
		input_ys.append(input_y_2)

	
	vec_x_mean = (input_xs[-1] * x1_mask[:, :, None]).sum(axis = 0) / x1_mask.sum(axis = 0)[:, None]
	vec_x_max = (input_xs[-1] * x1_mask[:, :, None]).max(axis = 0)
	
	vec_y_mean = (input_ys[-1] * x2_mask[:, :, None]).sum(axis = 0) / x2_mask.sum(axis = 0)[:, None]
	vec_y_max = (input_ys[-1] * x2_mask[:, :, None]).max(axis = 0)
	
	vec_x_y = tensor.concatenate([vec_x_mean, vec_x_max, vec_y_mean, vec_y_max], axis = -1)
	
	vec_x_y_dp = lib.dropout_layer(vec_x_y, use_noise, trng, options)

	vec_mlp = tensor.tanh(mlp_layer_3.build(tparams, vec_x_y_dp, options))
	
	vec_mlp_dp = lib.dropout_layer(vec_mlp, use_noise, trng, options)

	
	logit = mlp_layer_4.build(tparams, vec_mlp_dp, options)
	pred_prob = tensor.nnet.softmax(logit)
	pred = pred_prob.argmax(axis = -1)
	
	off = 1e-8
	if pred_prob.dtype == 'float32':
		off = 1e-6
	
	log_cost = -tensor.log(pred_prob[tensor.arange(x1.shape[1]), z] + off).mean()

	cost = log_cost

	if options.decay_c > 0.:
		decay_c = theano.shared(numpy.float32(options.decay_c), name='decay_c')
		weight_decay = 0.
		for kk, vv in tparams.items():
			weight_decay += (vv ** 2).sum()
		weight_decay *= decay_c
		cost += weight_decay

	# apply L2 regularization on W1
	if options.decay_w1 > 0.:
		decay_w1 = theano.shared(numpy.float32(options.decay_w1), name='decay_w1')
		weight_decay = 0.

		W1_list = split(tparams[_p('multihead_attention', 'W1')].get_value(), options.num_heads)

		for i in range(0, options.num_heads):
			for j in range(i + 1, options.num_heads):
				weight_decay += lib.ReLU(1 - ((W1_list[i] - W1_list[j]) ** 2).sum())
		weight_decay *= decay_w1
		cost += weight_decay

	if options.decay_rep > 0.:
		decay_rep = theano.shared(numpy.float32(options.decay_rep), name='decay_rep')
		weight_decay = 0.

		for i in range(0, options.num_heads):
			for j in range(i + 1, options.num_heads):
				weight_decay += lib.ReLU(
					1 - ((opt_ret['rep1_origin'][:, :, i] - opt_ret['rep1_origin'][:, :, j]) ** 2).sum(1).mean())
				weight_decay += lib.ReLU(
					1 - ((opt_ret['rep2_origin'][:, :, i] - opt_ret['rep2_origin'][:, :, j]) ** 2).sum(1).mean())

		weight_decay *= decay_rep
		cost += weight_decay

	if options.decay_att > 0.:
		decay_att = theano.shared(numpy.float32(options.decay_att), name='decay_att')
		weight_decay = 0.

		for i in range(0, options.num_heads):
			for j in range(i + 1, options.num_heads):
				weight_decay += lib.ReLU(
					1 - ((opt_ret['multiatt_1'][:, :, :, i] - opt_ret['multiatt_1'][:, :, :, j]) ** 2).sum(
						[0, 2]).mean())
				weight_decay += lib.ReLU(
					1 - ((opt_ret['multiatt_2'][:, :, :, i] - opt_ret['multiatt_2'][:, :, :, j]) ** 2).sum(
						[0, 2]).mean())

		weight_decay *= decay_att
		cost += weight_decay

	updated_params = OrderedDict([(key, value) for (key, value) in tparams.items() if not key.startswith('Wemb')])
	grads = tensor.grad(cost, wrt=itemlist(updated_params),disconnected_inputs='warn')
	# grads = tensor.grad(cost, wrt=tparams.values())

	if options.clip_c > 0.:
		g2 = 0.
		for g in grads:
			g2 += (g**2).sum()
		grad_norm = tensor.sqrt(g2)
		new_grads = []
		for g in grads:
			new_grads.append(tensor.switch(g2 > (options.clip_c ** 2), g * options.clip_c / tensor.sqrt(g2), g))
		grads = new_grads

	return params, updated_params,tparams, use_noise, x1, x1_mask,char_x1, char_x1_mask, x2, x2_mask, char_x2, char_x2_mask, z, pred_prob, pred, log_cost, cost, grad_norm, grads,opt_ret
	
def build_optimizer(lr, updated_params, x, y,  mask_x,mask_y, char_x1, char_x2,char_x1_mask, char_x2_mask, z, pred_prob, pred, log_cost, cost, grad_norm, grads, options):
	inps = [x, y,  mask_x,mask_y,char_x1, char_x2,char_x1_mask, char_x2_mask, z]
	f_grad_shared, f_update = optimizer[options.optimizer](lr, updated_params, inps, cost, grads)
	f_log_cost = theano.function(inputs = [ x, y,  mask_x,mask_y,char_x1, char_x2,char_x1_mask, char_x2_mask, z], outputs =log_cost, name = "f_log_cost")
	f_grad_norm = theano.function(inputs = [ x, y, mask_x,mask_y,char_x1, char_x2, char_x1_mask, char_x2_mask, z], outputs = grad_norm, name = "f_grad_norm")
	f_pred_prob = theano.function(inputs = [ x, y, mask_x,mask_y,char_x1, char_x2, char_x1_mask, char_x2_mask], outputs = pred_prob, name = "f_pred_prob")
	f_pred = theano.function(inputs = [ x, y,  mask_x,mask_y,char_x1, char_x2,char_x1_mask, char_x2_mask], outputs = pred, name = "f_pred")
	return f_grad_shared, f_update, f_log_cost, f_grad_norm, f_pred_prob, f_pred
